@@ -13,20 +13,23 @@ import numpy as np
 from binarycpython import _binary_c_bindings
 
 from binarycpython.utils.functions import (
-    binarycDecoder,
     temp_dir,
-    inspect_dict,
-    merge_dicts,
-    handle_ensemble_string_to_json,
     verbose_print,
-    extract_ensemble_json_from_string,
     is_capsule,
     Capturing,
 )
+from binarycpython.utils.ensemble import (
+    binarycDecoder,
+    handle_ensemble_string_to_json,
+    extract_ensemble_json_from_string,
+)
+from binarycpython.utils.dicts import (
+    inspect_dict,
+    merge_dicts,
+)
 
-# https://docs.python.org/3/library/unittest.html
-TMP_DIR = temp_dir()
-os.makedirs(os.path.join(TMP_DIR, "test"), exist_ok=True)
+
+TMP_DIR = temp_dir("tests", "test_c_bindings")
 
 #### some useful functions
 def return_argstring(
