@@ -3,7 +3,6 @@ Test cases for the grid
 
 Tasks:
     TODO: write tests for load_from_sourcefile
-    TODO: Set the temp_dir to grid_tests subdir. Should use that as an argument for the temp_dir() function
 """
 
 import os
@@ -13,19 +12,27 @@ import unittest
 import numpy as np
 
 from binarycpython.utils.grid import Population
+
 from binarycpython.utils.functions import (
     temp_dir,
-    extract_ensemble_json_from_string,
-    merge_dicts,
     remove_file,
     Capturing,
+    bin_data,
 )
+
+from binarycpython.utils.ensemble import (
+    extract_ensemble_json_from_string,
+)
+from binarycpython.utils.dicts import (
+    merge_dicts,
+)
+
 from binarycpython.utils.custom_logging_functions import binary_c_log_code
 
-binary_c_temp_dir = temp_dir()
-
+TMP_DIR = temp_dir("tests", "test_grid")
 TEST_VERBOSITY = 1
 
+
 def parse_function_test_grid_evolve_2_threads_with_custom_logging(self, output):
     """
     Simple parse function that directly appends all the output to a file
@@ -102,7 +109,7 @@ class test_Population(unittest.TestCase):
         """
 
         test_pop = Population()
-        test_pop.set(amt_cores=2, verbosity=TEST_VERBOSITY)
+        test_pop.set(num_cores=2, verbosity=TEST_VERBOSITY)
         test_pop.set(M_1=10)
         test_pop.set(data_dir="/tmp/binary_c_python")
         test_pop.set(ensemble_filter_SUPERNOVAE=1, ensemble_dt=1000)
@@ -115,7 +122,7 @@ class test_Population(unittest.TestCase):
         self.assertTrue(test_pop.bse_options["ensemble_filter_SUPERNOVAE"] == 1)
 
         #
-        self.assertTrue(test_pop.grid_options["amt_cores"] == 2)
+        self.assertTrue(test_pop.grid_options["num_cores"] == 2)
 
     def test_cmdline(self):
         with Capturing() as output:
@@ -132,8 +139,9 @@ class test_Population(unittest.TestCase):
         # make a dummy cmdline arg input
         sys.argv = [
             "script",
-            "--cmdline",
-            "metallicity=0.0002 amt_cores=2 data_dir=/tmp/binary_c_python",
+            "metallicity=0.0002",
+            "num_cores=2",
+            "data_dir=/tmp/binary_c_python",
         ]
 
         # Set up population
@@ -148,8 +156,8 @@ class test_Population(unittest.TestCase):
         self.assertTrue(test_pop.bse_options["metallicity"] == "0.0002")
 
         # Amt cores
-        self.assertTrue(isinstance(test_pop.grid_options["amt_cores"], int))
-        self.assertTrue(test_pop.grid_options["amt_cores"] == 2)
+        self.assertTrue(isinstance(test_pop.grid_options["num_cores"], int))
+        self.assertTrue(test_pop.grid_options["num_cores"] == 2)
 
         # datadir
         self.assertTrue(isinstance(test_pop.custom_options["data_dir"], str))
@@ -202,8 +210,7 @@ class test_Population(unittest.TestCase):
             name="lnm1",
             longname="Primary mass",
             valuerange=[1, 100],
-            resolution="{}".format(resolution["M_1"]),
-            spacingfunc="const(math.log(1), math.log(100), {})".format(
+            samplerfunc="const(math.log(1), math.log(100), {})".format(
                 resolution["M_1"]
             ),
             precode="M_1=math.exp(lnm1)",
@@ -217,8 +224,7 @@ class test_Population(unittest.TestCase):
             name="q",
             longname="Mass ratio",
             valuerange=["0.1/M_1", 1],
-            resolution="{}".format(resolution["q"]),
-            spacingfunc="const(0.1/M_1, 1, {})".format(resolution["q"]),
+            samplerfunc="const(0.1/M_1, 1, {})".format(resolution["q"]),
             probdist="flatsections(q, [{'min': 0.1/M_1, 'max': 1.0, 'height': 1}])",
             dphasevol="dq",
             precode="M_2 = q * M_1",
@@ -242,7 +248,7 @@ class test_Population(unittest.TestCase):
         test_pop = Population()
         test_pop.set(metallicity=0.02, verbosity=TEST_VERBOSITY)
         test_pop.set(M_1=10)
-        test_pop.set(amt_cores=2)
+        test_pop.set(num_cores=2)
         test_pop.set(data_dir="/tmp")
 
         population_settings = test_pop.return_population_settings()
@@ -252,22 +258,22 @@ class test_Population(unittest.TestCase):
         self.assertTrue(population_settings["bse_options"]["M_1"] == 10)
 
         self.assertIn("grid_options", population_settings)
-        self.assertTrue(population_settings["grid_options"]["amt_cores"] == 2)
+        self.assertTrue(population_settings["grid_options"]["num_cores"] == 2)
 
         self.assertIn("custom_options", population_settings)
         self.assertTrue(population_settings["custom_options"]["data_dir"] == "/tmp")
 
-    def test__return_binary_c_version_info(self):
+    def test_return_binary_c_version_info(self):
         with Capturing() as output:
-            self._test__return_binary_c_version_info()
+            self._test_return_binary_c_version_info()
 
-    def _test__return_binary_c_version_info(self):
+    def _test_return_binary_c_version_info(self):
         """
-        Unittests for the function _return_binary_c_version_info
+        Unittests for the function return_binary_c_version_info
         """
 
         test_pop = Population()
-        binary_c_version_info = test_pop._return_binary_c_version_info(parsed=True)
+        binary_c_version_info = test_pop.return_binary_c_version_info(parsed=True)
 
         self.assertTrue(isinstance(binary_c_version_info, dict))
         self.assertIn("isotopes", binary_c_version_info)
@@ -290,17 +296,17 @@ class test_Population(unittest.TestCase):
             if binary_c_version_info["macros"]["NUCSYN_ID_SOURCES"] == "on":
                 self.assertIsNotNone(binary_c_version_info["nucleosynthesis_sources"])
 
-    def test__return_binary_c_defaults(self):
+    def test_return_binary_c_defaults(self):
         with Capturing() as output:
-            self._test__return_binary_c_defaults()
+            self._test_return_binary_c_defaults()
 
-    def _test__return_binary_c_defaults(self):
+    def _test_return_binary_c_defaults(self):
         """
-        Unittests for the function _return_binary_c_defaults
+        Unittests for the function return_binary_c_defaults
         """
 
         test_pop = Population()
-        binary_c_defaults = test_pop._return_binary_c_defaults()
+        binary_c_defaults = test_pop.return_binary_c_defaults()
         self.assertIn("probability", binary_c_defaults)
         self.assertIn("phasevol", binary_c_defaults)
         self.assertIn("metallicity", binary_c_defaults)
@@ -341,8 +347,8 @@ class test_Population(unittest.TestCase):
 
         test_pop.set(metallicity=0.02, verbosity=TEST_VERBOSITY)
         test_pop.set(M_1=10)
-        test_pop.set(amt_cores=2)
-        test_pop.set(data_dir=binary_c_temp_dir)
+        test_pop.set(num_cores=2)
+        test_pop.set(data_dir=TMP_DIR)
 
         # datadir
         settings_filename = test_pop.export_all_info(use_datadir=True)
@@ -366,7 +372,7 @@ class test_Population(unittest.TestCase):
         # datadir
         settings_filename = test_pop.export_all_info(
             use_datadir=False,
-            outfile=os.path.join(binary_c_temp_dir, "example_settings.json"),
+            outfile=os.path.join(TMP_DIR, "example_settings.json"),
         )
         self.assertTrue(os.path.isfile(settings_filename))
         with open(settings_filename, "r") as f:
@@ -389,7 +395,7 @@ class test_Population(unittest.TestCase):
             ValueError,
             test_pop.export_all_info,
             use_datadir=False,
-            outfile=os.path.join(binary_c_temp_dir, "example_settings.txt"),
+            outfile=os.path.join(TMP_DIR, "example_settings.txt"),
         )
 
     def test__cleanup_defaults(self):
@@ -440,7 +446,7 @@ class test_Population(unittest.TestCase):
         Unittests for the function _dict_from_line_source_file
         """
 
-        source_file = os.path.join(binary_c_temp_dir, "example_source_file.txt")
+        source_file = os.path.join(TMP_DIR, "example_source_file.txt")
 
         # write
         with open(source_file, "w") as f:
@@ -469,7 +475,7 @@ class test_Population(unittest.TestCase):
         Printf("TEST_CUSTOM_LOGGING_1 %30.12e %g %g %g %g\\n",
             //
             stardata->model.time, // 1
-            
+
             // masses
             stardata->common.zero_age.mass[0], //
             stardata->common.zero_age.mass[1], //
@@ -485,7 +491,8 @@ class test_Population(unittest.TestCase):
             M_2=5,
             orbital_period=100000,
             metallicty=0.02,
-            max_evolution_time=15000, verbosity=TEST_VERBOSITY
+            max_evolution_time=15000,
+            verbosity=TEST_VERBOSITY,
         )
 
         test_pop.set(C_logging_code=CUSTOM_LOGGING_STRING_MASSES)
@@ -504,7 +511,8 @@ class test_Population(unittest.TestCase):
             M_2=5,
             orbital_period=100000,
             metallicty=0.02,
-            max_evolution_time=15000, verbosity=TEST_VERBOSITY
+            max_evolution_time=15000,
+            verbosity=TEST_VERBOSITY,
         )
 
         test_pop_2.set(C_auto_logging=custom_logging_dict)
@@ -532,7 +540,7 @@ class test_grid_evolve(unittest.TestCase):
 
         test_pop_evolve_1_thread = Population()
         test_pop_evolve_1_thread.set(
-            amt_cores=1, M_2=1, orbital_period=100000, verbosity=TEST_VERBOSITY
+            num_cores=1, M_2=1, orbital_period=100000, verbosity=TEST_VERBOSITY
         )
 
         resolution = {"M_1": 10}
@@ -541,8 +549,7 @@ class test_grid_evolve(unittest.TestCase):
             name="lnm1",
             longname="Primary mass",
             valuerange=[1, 100],
-            resolution="{}".format(resolution["M_1"]),
-            spacingfunc="const(math.log(1), math.log(100), {})".format(
+            samplerfunc="const(math.log(1), math.log(100), {})".format(
                 resolution["M_1"]
             ),
             precode="M_1=math.exp(lnm1)",
@@ -554,7 +561,9 @@ class test_grid_evolve(unittest.TestCase):
 
         analytics = test_pop_evolve_1_thread.evolve()
         self.assertLess(
-            np.abs(analytics["total_probability"] - 0.10820655287892997), 1e-10, msg=analytics["total_probability"]
+            np.abs(analytics["total_probability"] - 0.10820655287892997),
+            1e-10,
+            msg=analytics["total_probability"],
         )
         self.assertTrue(analytics["total_count"] == 10)
 
@@ -568,7 +577,9 @@ class test_grid_evolve(unittest.TestCase):
         """
 
         test_pop = Population()
-        test_pop.set(amt_cores=2, M_2=1, orbital_period=100000, verbosity=TEST_VERBOSITY)
+        test_pop.set(
+            num_cores=2, M_2=1, orbital_period=100000, verbosity=TEST_VERBOSITY
+        )
 
         resolution = {"M_1": 10}
 
@@ -576,8 +587,7 @@ class test_grid_evolve(unittest.TestCase):
             name="lnm1",
             longname="Primary mass",
             valuerange=[1, 100],
-            resolution="{}".format(resolution["M_1"]),
-            spacingfunc="const(math.log(1), math.log(100), {})".format(
+            samplerfunc="const(math.log(1), math.log(100), {})".format(
                 resolution["M_1"]
             ),
             precode="M_1=math.exp(lnm1)",
@@ -589,7 +599,9 @@ class test_grid_evolve(unittest.TestCase):
 
         analytics = test_pop.evolve()
         self.assertLess(
-            np.abs(analytics["total_probability"] - 0.10820655287892997), 1e-10, msg=analytics["total_probability"]
+            np.abs(analytics["total_probability"] - 0.10820655287892997),
+            1e-10,
+            msg=analytics["total_probability"],
         )  #
         self.assertTrue(analytics["total_count"] == 10)
 
@@ -602,21 +614,20 @@ class test_grid_evolve(unittest.TestCase):
         Unittests to see if multiple threads do the custom logging correctly
         """
 
-        data_dir_value = os.path.join(binary_c_temp_dir, "grid_tests")
-        amt_cores_value = 2
+        data_dir_value = os.path.join(TMP_DIR, "grid_tests")
+        num_cores_value = 2
         custom_logging_string = 'Printf("MY_STELLAR_DATA_TEST_EXAMPLE %g %g %g %g\\n",((double)stardata->model.time),((double)stardata->star[0].mass),((double)stardata->model.probability),((double)stardata->model.dt));'
 
         test_pop = Population()
 
         test_pop.set(
-            amt_cores=amt_cores_value,
+            num_cores=num_cores_value,
             verbosity=TEST_VERBOSITY,
             M_2=1,
             orbital_period=100000,
             data_dir=data_dir_value,
             C_logging_code=custom_logging_string,  # input it like this.
             parse_function=parse_function_test_grid_evolve_2_threads_with_custom_logging,
-
         )
         test_pop.set(ensemble=0)
         resolution = {"M_1": 2}
@@ -625,8 +636,7 @@ class test_grid_evolve(unittest.TestCase):
             name="lnm1",
             longname="Primary mass",
             valuerange=[1, 100],
-            resolution="{}".format(resolution["M_1"]),
-            spacingfunc="const(math.log(1), math.log(100), {})".format(
+            samplerfunc="const(math.log(1), math.log(100), {})".format(
                 resolution["M_1"]
             ),
             precode="M_1=math.exp(lnm1)",
@@ -644,7 +654,7 @@ class test_grid_evolve(unittest.TestCase):
                     analytics["population_name"], thread_id
                 ),
             )
-            for thread_id in range(amt_cores_value)
+            for thread_id in range(num_cores_value)
         ]
 
         for output_name in output_names:
@@ -667,7 +677,9 @@ class test_grid_evolve(unittest.TestCase):
         """
 
         test_pop = Population()
-        test_pop.set(amt_cores=2, M_2=1, orbital_period=100000, verbosity=TEST_VERBOSITY)
+        test_pop.set(
+            num_cores=2, M_2=1, orbital_period=100000, verbosity=TEST_VERBOSITY
+        )
 
         # Set the amt of failed systems that each thread will log
         test_pop.set(failed_systems_threshold=4)
@@ -677,7 +689,7 @@ Exit_binary_c(BINARY_C_NORMAL_EXIT, "testing exits. This is part of the testing,
 Printf("TEST_CUSTOM_LOGGING_1 %30.12e %g %g %g %g\\n",
     //
     stardata->model.time, // 1
-    
+
     // masses
     stardata->common.zero_age.mass[0], //
     stardata->common.zero_age.mass[1], //
@@ -694,8 +706,7 @@ Printf("TEST_CUSTOM_LOGGING_1 %30.12e %g %g %g %g\\n",
             name="lnm1",
             longname="Primary mass",
             valuerange=[1, 100],
-            resolution="{}".format(resolution["M_1"]),
-            spacingfunc="const(math.log(1), math.log(100), {})".format(
+            samplerfunc="const(math.log(1), math.log(100), {})".format(
                 resolution["M_1"]
             ),
             precode="M_1=math.exp(lnm1)",
@@ -707,7 +718,9 @@ Printf("TEST_CUSTOM_LOGGING_1 %30.12e %g %g %g %g\\n",
 
         analytics = test_pop.evolve()
         self.assertLess(
-            np.abs(analytics["total_probability"] - 0.10820655287892997), 1e-10, msg=analytics["total_probability"]
+            np.abs(analytics["total_probability"] - 0.10820655287892997),
+            1e-10,
+            msg=analytics["total_probability"],
         )  #
         self.assertEqual(analytics["failed_systems_error_codes"], [0])
         self.assertTrue(analytics["total_count"] == 10)
@@ -718,7 +731,9 @@ Printf("TEST_CUSTOM_LOGGING_1 %30.12e %g %g %g %g\\n",
         # test to see if 1 thread does all the systems
 
         test_pop = Population()
-        test_pop.set(amt_cores=2, M_2=1, orbital_period=100000, verbosity=TEST_VERBOSITY)
+        test_pop.set(
+            num_cores=2, M_2=1, orbital_period=100000, verbosity=TEST_VERBOSITY
+        )
         test_pop.set(failed_systems_threshold=4)
         test_pop.set(C_logging_code=CUSTOM_LOGGING_STRING_WITH_EXIT)
 
@@ -728,8 +743,7 @@ Printf("TEST_CUSTOM_LOGGING_1 %30.12e %g %g %g %g\\n",
             name="lnm1",
             longname="Primary mass",
             valuerange=[1, 100],
-            resolution="{}".format(resolution["M_1"]),
-            spacingfunc="const(math.log(1), math.log(100), {})".format(
+            samplerfunc="const(math.log(1), math.log(100), {})".format(
                 resolution["M_1"]
             ),
             precode="M_1=math.exp(lnm1)",
@@ -743,8 +757,7 @@ Printf("TEST_CUSTOM_LOGGING_1 %30.12e %g %g %g %g\\n",
             name="q",
             longname="Mass ratio",
             valuerange=["0.1/M_1", 1],
-            resolution="{}".format(resolution["q"]),
-            spacingfunc="const(0.1/M_1, 1, {})".format(resolution["q"]),
+            samplerfunc="const(0.1/M_1, 1, {})".format(resolution["q"]),
             probdist="flatsections(q, [{'min': 0.1/M_1, 'max': 1.0, 'height': 1}])",
             dphasevol="dq",
             precode="M_2 = q * M_1",
@@ -770,7 +783,9 @@ Printf("TEST_CUSTOM_LOGGING_1 %30.12e %g %g %g %g\\n",
         """
 
         test_pop = Population()
-        test_pop.set(amt_cores=1, M_2=1, orbital_period=100000, verbosity=TEST_VERBOSITY)
+        test_pop.set(
+            num_cores=1, M_2=1, orbital_period=100000, verbosity=TEST_VERBOSITY
+        )
 
         resolution = {"M_1": 10}
         self.assertRaises(ValueError, test_pop.evolve)
@@ -784,12 +799,12 @@ Printf("TEST_CUSTOM_LOGGING_1 %30.12e %g %g %g %g\\n",
         Unittests to see if multiple threads output the ensemble information to files correctly
         """
 
-        data_dir_value = binary_c_temp_dir
-        amt_cores_value = 2
+        data_dir_value = TMP_DIR
+        num_cores_value = 2
 
         test_pop = Population()
         test_pop.set(
-            amt_cores=amt_cores_value,
+            num_cores=num_cores_value,
             verbosity=TEST_VERBOSITY,
             M_2=1,
             orbital_period=100000,
@@ -800,7 +815,7 @@ Printf("TEST_CUSTOM_LOGGING_1 %30.12e %g %g %g %g\\n",
             ensemble_dt=1000,
         )
         test_pop.set(
-            data_dir=binary_c_temp_dir,
+            data_dir=TMP_DIR,
             ensemble_output_name="ensemble_output.json",
             combine_ensemble_with_thread_joining=False,
         )
@@ -811,8 +826,7 @@ Printf("TEST_CUSTOM_LOGGING_1 %30.12e %g %g %g %g\\n",
             name="lnm1",
             longname="Primary mass",
             valuerange=[1, 100],
-            resolution="{}".format(resolution["M_1"]),
-            spacingfunc="const(math.log(1), math.log(100), {})".format(
+            samplerfunc="const(math.log(1), math.log(100), {})".format(
                 resolution["M_1"]
             ),
             precode="M_1=math.exp(lnm1)",
@@ -830,7 +844,7 @@ Printf("TEST_CUSTOM_LOGGING_1 %30.12e %g %g %g %g\\n",
                     analytics["population_name"], thread_id
                 ),
             )
-            for thread_id in range(amt_cores_value)
+            for thread_id in range(num_cores_value)
         ]
 
         for output_name in output_names:
@@ -839,9 +853,7 @@ Printf("TEST_CUSTOM_LOGGING_1 %30.12e %g %g %g %g\\n",
             with open(output_name, "r") as f:
                 file_content = f.read()
 
-                self.assertTrue(file_content.startswith("ENSEMBLE_JSON"))
-
-                ensemble_json = extract_ensemble_json_from_string(file_content)
+                ensemble_json = json.loads(file_content)
 
                 self.assertTrue(isinstance(ensemble_json, dict))
                 self.assertNotEqual(ensemble_json, {})
@@ -858,12 +870,12 @@ Printf("TEST_CUSTOM_LOGGING_1 %30.12e %g %g %g %g\\n",
         Unittests to see if multiple threads correclty combine the ensemble data and store them in the grid
         """
 
-        data_dir_value = binary_c_temp_dir
-        amt_cores_value = 2
+        data_dir_value = TMP_DIR
+        num_cores_value = 2
 
         test_pop = Population()
         test_pop.set(
-            amt_cores=amt_cores_value,
+            num_cores=num_cores_value,
             verbosity=TEST_VERBOSITY,
             M_2=1,
             orbital_period=100000,
@@ -874,7 +886,7 @@ Printf("TEST_CUSTOM_LOGGING_1 %30.12e %g %g %g %g\\n",
             ensemble_dt=1000,
         )
         test_pop.set(
-            data_dir=binary_c_temp_dir,
+            data_dir=TMP_DIR,
             combine_ensemble_with_thread_joining=True,
             ensemble_output_name="ensemble_output.json",
         )
@@ -885,8 +897,7 @@ Printf("TEST_CUSTOM_LOGGING_1 %30.12e %g %g %g %g\\n",
             name="lnm1",
             longname="Primary mass",
             valuerange=[1, 100],
-            resolution="{}".format(resolution["M_1"]),
-            spacingfunc="const(math.log(1), math.log(100), {})".format(
+            samplerfunc="const(math.log(1), math.log(100), {})".format(
                 resolution["M_1"]
             ),
             precode="M_1=math.exp(lnm1)",
@@ -915,13 +926,13 @@ Printf("TEST_CUSTOM_LOGGING_1 %30.12e %g %g %g %g\\n",
         Unittests to compare the method of storing the combined ensemble data in the object and writing them to files and combining them later. they have to be the same
         """
 
-        data_dir_value = binary_c_temp_dir
-        amt_cores_value = 2
+        data_dir_value = TMP_DIR
+        num_cores_value = 2
 
         # First
         test_pop_1 = Population()
         test_pop_1.set(
-            amt_cores=amt_cores_value,
+            num_cores=num_cores_value,
             verbosity=TEST_VERBOSITY,
             M_2=1,
             orbital_period=100000,
@@ -932,7 +943,7 @@ Printf("TEST_CUSTOM_LOGGING_1 %30.12e %g %g %g %g\\n",
             ensemble_dt=1000,
         )
         test_pop_1.set(
-            data_dir=binary_c_temp_dir,
+            data_dir=TMP_DIR,
             combine_ensemble_with_thread_joining=True,
             ensemble_output_name="ensemble_output.json",
         )
@@ -943,8 +954,7 @@ Printf("TEST_CUSTOM_LOGGING_1 %30.12e %g %g %g %g\\n",
             name="lnm1",
             longname="Primary mass",
             valuerange=[1, 100],
-            resolution="{}".format(resolution["M_1"]),
-            spacingfunc="const(math.log(1), math.log(100), {})".format(
+            samplerfunc="const(math.log(1), math.log(100), {})".format(
                 resolution["M_1"]
             ),
             precode="M_1=math.exp(lnm1)",
@@ -960,7 +970,7 @@ Printf("TEST_CUSTOM_LOGGING_1 %30.12e %g %g %g %g\\n",
         # second
         test_pop_2 = Population()
         test_pop_2.set(
-            amt_cores=amt_cores_value,
+            num_cores=num_cores_value,
             verbosity=TEST_VERBOSITY,
             M_2=1,
             orbital_period=100000,
@@ -971,7 +981,7 @@ Printf("TEST_CUSTOM_LOGGING_1 %30.12e %g %g %g %g\\n",
             ensemble_dt=1000,
         )
         test_pop_2.set(
-            data_dir=binary_c_temp_dir,
+            data_dir=TMP_DIR,
             ensemble_output_name="ensemble_output.json",
             combine_ensemble_with_thread_joining=False,
         )
@@ -982,8 +992,7 @@ Printf("TEST_CUSTOM_LOGGING_1 %30.12e %g %g %g %g\\n",
             name="lnm1",
             longname="Primary mass",
             valuerange=[1, 100],
-            resolution="{}".format(resolution["M_1"]),
-            spacingfunc="const(math.log(1), math.log(100), {})".format(
+            samplerfunc="const(math.log(1), math.log(100), {})".format(
                 resolution["M_1"]
             ),
             precode="M_1=math.exp(lnm1)",
@@ -1001,7 +1010,7 @@ Printf("TEST_CUSTOM_LOGGING_1 %30.12e %g %g %g %g\\n",
                     analytics_2["population_name"], thread_id
                 ),
             )
-            for thread_id in range(amt_cores_value)
+            for thread_id in range(num_cores_value)
         ]
         ensemble_output_2 = {}
 
@@ -1011,9 +1020,7 @@ Printf("TEST_CUSTOM_LOGGING_1 %30.12e %g %g %g %g\\n",
             with open(output_name, "r") as f:
                 file_content = f.read()
 
-                self.assertTrue(file_content.startswith("ENSEMBLE_JSON"))
-
-                ensemble_json = extract_ensemble_json_from_string(file_content)
+                ensemble_json = json.loads(file_content)
 
                 ensemble_output_2 = merge_dicts(ensemble_output_2, ensemble_json)
 
@@ -1032,5 +1039,146 @@ Printf("TEST_CUSTOM_LOGGING_1 %30.12e %g %g %g %g\\n",
             )
 
 
+def parse_function_adding_results(self, output):
+    """
+    Example parse function
+    """
+
+    seperator = " "
+
+    parameters = ["time", "mass", "zams_mass", "probability", "stellar_type"]
+
+    self.grid_results["example"]["count"] += 1
+
+    # Go over the output.
+    for line in output.splitlines():
+        headerline = line.split()[0]
+
+        # CHeck the header and act accordingly
+        if headerline == "EXAMPLE_OUTPUT":
+            values = line.split()[1:]
+
+            # Bin the mass probability
+            self.grid_results["example"]["mass"][
+                bin_data(float(values[2]), binwidth=0.5)
+            ] += float(values[3])
+
+            #
+            if not len(parameters) == len(values):
+                print("Number of column names isnt equal to number of columns")
+                raise ValueError
+
+    # record the probability of this line (Beware, this is meant to only be run once for each system. its a controls quantity)
+    self.grid_results["example"]["probability"] += float(values[3])
+
+
+class test_resultdict(unittest.TestCase):
+    """
+    Unittests for bin_data
+    """
+
+    def test_adding_results(self):
+        """
+        Function to test whether the results are properly added and combined
+        """
+
+        # Create custom logging statement
+        custom_logging_statement = """
+        if (stardata->model.time < stardata->model.max_evolution_time)
+        {
+            Printf("EXAMPLE_OUTPUT %30.16e %g %g %30.12e %d\\n",
+                //
+                stardata->model.time, // 1
+                stardata->star[0].mass, // 2
+                stardata->common.zero_age.mass[0], // 3
+                stardata->model.probability, // 4
+                stardata->star[0].stellar_type // 5
+          );
+        };
+        /* Kill the simulation to save time */
+        stardata->model.max_evolution_time = stardata->model.time - stardata->model.dtm;
+        """
+
+        example_pop = Population()
+        example_pop.set(verbosity=0)
+        example_pop.set(
+            max_evolution_time=15000,  # bse_options
+            # grid_options
+            num_cores=3,
+            tmp_dir=TMP_DIR,
+            # Custom options
+            data_dir=os.path.join(TMP_DIR, "test_resultdict"),  # custom_options
+            C_logging_code=custom_logging_statement,
+            parse_function=parse_function_adding_results,
+        )
+
+        # Add grid variables
+        resolution = {"M_1": 10}
+
+        # Mass
+        example_pop.add_grid_variable(
+            name="lnm1",
+            longname="Primary mass",
+            valuerange=[2, 150],
+            samplerfunc="const(math.log(2), math.log(150), {})".format(
+                resolution["M_1"]
+            ),
+            precode="M_1=math.exp(lnm1)",
+            probdist="three_part_powerlaw(M_1, 0.1, 0.5, 1.0, 150, -1.3, -2.3, -2.3)*M_1",
+            dphasevol="dlnm1",
+            parameter_name="M_1",
+            condition="",  # Impose a condition on this grid variable. Mostly for a check for yourself
+        )
+
+        ## Executing a population
+        ## This uses the values generated by the grid_variables
+        analytics = example_pop.evolve()
+
+        #
+        grid_prob = analytics["total_probability"]
+        result_dict_prob = example_pop.grid_results["example"]["probability"]
+
+        # amt systems
+        grid_count = analytics["total_count"]
+        result_dict_count = example_pop.grid_results["example"]["count"]
+
+        # Check if the total probability matches
+        self.assertAlmostEqual(
+            grid_prob,
+            result_dict_prob,
+            places=12,
+            msg="Total probability from grid {} and from result dict {} are not equal".format(
+                grid_prob, result_dict_prob
+            ),
+        )
+
+        # Check if the total count matches
+        self.assertEqual(
+            grid_count,
+            result_dict_count,
+            msg="Total count from grid {} and from result dict {} are not equal".format(
+                grid_count, result_dict_count
+            ),
+        )
+
+        # Check if the structure is what we expect. Note: this depends on the probability calculation. if that changes we need to recalibrate this
+        test_case_dict = {
+            2.25: 0.01895481306515,
+            3.75: 0.01081338190204,
+            5.75: 0.006168841009268,
+            9.25: 0.003519213484031,
+            13.75: 0.002007648361756,
+            21.25: 0.001145327489437,
+            33.25: 0.0006533888518775,
+            50.75: 0.0003727466560393,
+            78.25: 0.000212645301782,
+            120.75: 0.0001213103421247,
+        }
+
+        self.assertEqual(
+            test_case_dict, dict(example_pop.grid_results["example"]["mass"])
+        )
+
+
 if __name__ == "__main__":
     unittest.main()
