@@ -14,7 +14,7 @@ from binarycpython.utils.grid_options_defaults import (
     grid_options_description_checker,
 )
 
-binary_c_temp_dir = temp_dir()
+TMP_DIR = temp_dir("tests", "test_grid_options_defaults")
 
 
 class test_grid_options_defaults(unittest.TestCase):
@@ -35,7 +35,7 @@ class test_grid_options_defaults(unittest.TestCase):
         result_1 = grid_options_help(input_1)
         self.assertEqual(result_1, {}, msg="Dict should be empty")
 
-        input_2 = "amt_cores"
+        input_2 = "num_cores"
         result_2 = grid_options_help(input_2)
         self.assertIn(
             input_2,
@@ -78,15 +78,11 @@ class test_grid_options_defaults(unittest.TestCase):
         Unit tests for the grid_options_description_checker function
         """
 
-        input_1 = os.path.join(
-            binary_c_temp_dir, "test_write_grid_options_to_rst_file_1.txt"
-        )
+        input_1 = os.path.join(TMP_DIR, "test_write_grid_options_to_rst_file_1.txt")
         output_1 = write_grid_options_to_rst_file(input_1)
         self.assertIsNone(output_1)
 
-        input_2 = os.path.join(
-            binary_c_temp_dir, "test_write_grid_options_to_rst_file_2.rst"
-        )
+        input_2 = os.path.join(TMP_DIR, "test_write_grid_options_to_rst_file_2.rst")
         output_2 = write_grid_options_to_rst_file(input_2)
 
         self.assertTrue(os.path.isfile(input_2))
