@@ -8,46 +8,189 @@ Tasks:
     - TODO: change all prints to verbose_prints
 """
 
-import json
-import os
-import tempfile
+
+import bz2
+import collections
+from colorama import Fore, Back, Style
 import copy
+import datetime as dt
+import gc
+import gzip
+from halo import Halo
+import h5py
+import humanize
 import inspect
+from io import StringIO
+import json
+import msgpack
+import numpy as np
+import os
+import psutil
+import py_rinterpolate
+import re
+import resource
 import sys
 import subprocess
+import tempfile
 import time
 import types
-
-from io import StringIO
 from typing import Union, Any
-from collections import (
-    defaultdict,
-    OrderedDict,
-)
 
-import h5py
-import numpy as np
+import simplejson
 
-from binarycpython import _binary_c_bindings
+# import orjson
+
+import astropy.units as u
 import binarycpython.utils.moe_di_stefano_2017_data as moe_di_stefano_2017_data
 
-import py_rinterpolate
+from binarycpython import _binary_c_bindings
+from binarycpython.utils.dicts import filter_dict, filter_dict_through_values
+
 
 ########################################################
 # Unsorted
 ########################################################
+
+
+def format_number(number):
+    # compact number formatter
+
+    string = "{number:.2g}".format(number=number)
+    string = string.replace("e+0", "e+")
+    string = string.replace("e-0", "e-")
+
+    return string
+
+
+def check_if_in_shell():
+    """
+    Function to check whether the script is running from a shell
+    """
+
+    if sys.stdin and sys.stdin.isatty():
+        in_shell = True
+    else:
+        in_shell = False
+
+    return in_shell
+
+
+def timedelta(delta):
+    """
+    Function to convert a length of time (float, seconds) to a string for
+    human-readable output.
+    """
+    # currently use the humanize module to do this
+    t = humanize.time.precisedelta(
+        dt.timedelta(seconds=delta),
+        format="%0.2f",
+        minimum_unit="milliseconds",
+        suppress=["milliseconds"],
+    )
+    # and use more compact units
+    t = t.replace(" days and", "d")
+    t = t.replace(" hours and", "h")
+    t = t.replace(" minutes and", "m")
+    t = t.replace(" seconds and", "s")
+    t = t.replace(" days", "d")
+    t = t.replace(" hours", "h")
+    t = t.replace(" minutes", "m")
+    t = t.replace(" seconds", "s")
+    return t
+
+
+def get_ANSI_colours():
+    # ANSI colours dictionary
+    foreground_colours = {
+        "red": Fore.RED,
+        "yellow": Fore.YELLOW,
+        "blue": Fore.BLUE,
+        "cyan": Fore.CYAN,
+        "green": Fore.GREEN,
+        "magenta": Fore.MAGENTA,
+        "white": Fore.WHITE,
+        "black": Fore.BLACK,
+    }
+
+    background_colours = {
+        "red": Back.RED,
+        "yellow": Back.YELLOW,
+        "blue": Back.BLUE,
+        "cyan": Back.CYAN,
+        "green": Back.GREEN,
+        "magenta": Back.MAGENTA,
+        "white": Back.WHITE,
+        "black": Back.BLACK,
+    }
+
+    default_style = Style.BRIGHT
+    colours = {}
+
+    for c in foreground_colours:
+        colours[c] = default_style + foreground_colours[c]
+        for d in background_colours:
+            colours[c + " on " + d] = foreground_colours[c] + background_colours[d]
+    colours["reset"] = Style.RESET_ALL
+
+    return colours
+
+
+def mem_use():
+    """
+    Return current process memory use in MB. (Takes no arguments) Note: this is per-thread only.
+    """
+
+    return resource.getrusage(resource.RUSAGE_SELF).ru_maxrss / 1024.0
+
+
+def trem(dt, count, dn, n):
+    """
+    Estimate time remaining (seconds) given a differential time and count (i.e. progress = $count/$n). $dt is the time since the last call, $count is the current progress count, $dn is the number run since the last call, and $n is the total number required.
+    """
+    tpr = dt / max(1, dn)
+    etasecs = tpr * (n - count)
+    (eta, units) = conv_time_units(etasecs)
+    return (eta, units, tpr, etasecs)
+
+
+def conv_time_units(t):
+    """
+    Converts time (t, in seconds, passing in as the only argument) to seconds, minutes or hours depending on its magnitude. Returns a tuple (t,units).
+    """
+    units = "s"
+    # default to seconds
+    if t > 60:
+        t /= 60
+        units = "m"
+    if t > 60:
+        t /= 60
+        units = "h"
+    return (t, units)
+
+
+def bin_data(value, binwidth):
+    """
+    Function that bins the data
+
+    Uses the absolute value of binwidth
+    """
+
+    return ((0.5 if value > 0.0 else -0.5) + int(value / abs(binwidth))) * abs(binwidth)
+
+
 def convert_bytes(size):
     """
     Function to return the size + a magnitude string
     """
 
-    for name in ['bytes', 'KB', 'MB', 'GB', 'TB']:
+    for name in ["bytes", "KB", "MB", "GB", "TB"]:
         if size < 1024.0:
             return "%3.1f %s" % (size, name)
         size /= 1024.0
 
     return size
 
+
 def get_size(obj, seen=None):
     """
     Recursively finds size of objects
@@ -67,167 +210,28 @@ def get_size(obj, seen=None):
     if isinstance(obj, dict):
         size += sum([get_size(v, seen) for v in obj.values()])
         size += sum([get_size(k, seen) for k in obj.keys()])
-    elif hasattr(obj, '__dict__'):
+    elif hasattr(obj, "__dict__"):
         size += get_size(obj.__dict__, seen)
-    elif hasattr(obj, '__iter__') and not isinstance(obj, (str, bytes, bytearray)):
+    elif hasattr(obj, "__iter__") and not isinstance(obj, (str, bytes, bytearray)):
         size += sum([get_size(i, seen) for i in obj])
     return size
 
-def subtract_dicts(dict_1: dict, dict_2: dict) -> dict:
-    """
-    Function to subtract two dictionaries.
-
-    Only allows values to be either a dict or a numerical type
-
-    For the overlapping keys (key name present in both dicts):
-        When the keys are of the same type:
-            - If the types are of numerical type: subtract the value at dict 2 from dict 1.
-            - If the types are both dictionaries: call this function with the subdicts
-
-        When the keys are not of the same type:
-            - if the keys are all of numerical types
-
-    For the unique keys:
-        - if the key is from dict 1: adds the value to the new dict (be it numerical value or dict)
-        - If the key is from dict 2: Adds the negative of its value in case of numerical type.
-            if the type is a dict, the result of subtract_dicts({}, dict_2[key]) will be set
-
-    If the result is 0, the key will be removed from the resulting dict.
-    If that results in an empty dict, the dict will be removed too.
-
-    Args:
-        dict_1: first dictionary
-        dict_2: second dictionary
-
-    Returns:
-        Subtracted dictionary
-    """
-
-    # Set up new dict
-    new_dict = {}
-
-    # Define allowed numerical types
-    ALLOWED_NUMERICAL_TYPES = (float, int, np.float64)
-
-    #
-    keys_1 = dict_1.keys()
-    keys_2 = dict_2.keys()
-
-    # Find overlapping keys of both dicts
-    overlapping_keys = set(keys_1).intersection(set(keys_2))
-
-    # Find the keys that are unique
-    unique_to_dict_1 = set(keys_1).difference(set(keys_2))
-    unique_to_dict_2 = set(keys_2).difference(set(keys_1))
-
-    # Add the unique keys to the new dict
-    for key in unique_to_dict_1:
-        # If these items are numerical types
-        if isinstance(dict_1[key], ALLOWED_NUMERICAL_TYPES):
-            new_dict[key] = dict_1[key]
-            if new_dict[key] == 0:
-                del new_dict[key]
-
-        # Else, to be safe we should deepcopy them
-        elif isinstance(dict_1[key], dict):
-            copy_dict = copy.deepcopy(dict_1[key])
-            new_dict[key] = copy_dict
-        else:
-            msg = "Error: using unsupported type for key {}: {}".format(
-                key, type(dict_1[key])
-            )
-            print(msg)
-            raise ValueError(msg)
-
-    # Add the unique keys to the new dict
-    for key in unique_to_dict_2:
-        # If these items are numerical type, we should add the negative of the value
-        if isinstance(dict_2[key], ALLOWED_NUMERICAL_TYPES):
-            new_dict[key] = -dict_2[key]
-            if new_dict[key] == 0:
-                del new_dict[key]
-
-        # Else we should place the negative of that dictionary in the new place
-        elif isinstance(dict_2[key], dict):
-            new_dict[key] = subtract_dicts({}, dict_2[key])
-        else:
-            msg = "Error: using unsupported type for key {}: {}".format(
-                key, type(dict_2[key])
-            )
-            print(msg)
-            raise ValueError(msg)
-
-    # Go over the common keys:
-    for key in overlapping_keys:
-
-        # See whether the types are actually the same
-        if not type(dict_1[key]) is type(dict_2[key]):
-            # Exceptions:
-            if (type(dict_1[key]) in ALLOWED_NUMERICAL_TYPES) and (
-                type(dict_2[key]) in ALLOWED_NUMERICAL_TYPES
-            ):
-                # We can safely subtract the values since they are all numeric
-                new_dict[key] = dict_1[key] - dict_2[key]
-                if new_dict[key] == 0:
-                    del new_dict[key]
-
-            else:
-                msg = "Error key: {} value: {} type: {} and key: {} value: {} type: {} are not of the same type and cannot be merged".format(
-                    key,
-                    dict_1[key],
-                    type(dict_1[key]),
-                    key,
-                    dict_2[key],
-                    type(dict_2[key]),
-                )
-
-                print(msg)
-                raise ValueError(msg)
-
-        # This is where the keys are the same
-        else:
-            # If these items are numeric types
-            if isinstance(dict_1[key], ALLOWED_NUMERICAL_TYPES):
-                new_dict[key] = dict_1[key] - dict_2[key]
-
-                # Remove entry if the value is 0
-                if new_dict[key] == 0:
-                    del new_dict[key]
-
-            # Else, to be safe we should deepcopy them
-            elif isinstance(dict_1[key], dict):
-                new_dict[key] = subtract_dicts(dict_1[key], dict_2[key])
-
-                # Remove entry if it results in an empty dict
-                # TODO: write test to prevent empty dicts from showing up
-                if not new_dict[key]:
-                    del new_dict[key]
-            else:
-                msg = "Error: using unsupported type for key {}: {}".format(
-                    key, type(dict_2[key])
-                )
-                print(msg)
-                raise ValueError(msg)
-
-    #
-    return new_dict
-
 
 def get_moe_di_stefano_dataset(options, verbosity=0):
     """
-    Function to get the default moe and di Stefano dataset or accept a user input.
+    Function to get the default Moe and di Stefano dataset or accept a user input.
+
+    Returns a dict containing the (JSON) data.
     """
 
-    if not options.get("file", None):
-        verbose_print(
-            "Using the default Moe and de Stefano 2017 datafile",
-            verbosity,
-            1,
-        )
+    json_data = None
 
-        json_data = copy.deepcopy(moe_di_stefano_2017_data.moe_di_stefano_2017_data)
+    if "JSON" in options:
+        # use the JSON data passed in
+        json_data = options["JSON"]
 
-    else:
+    elif "file" in options:
+        # use the file passed in, if provided
         if not os.path.isfile(options["file"]):
             verbose_print(
                 "The provided 'file' Moe and de Stefano JSON file does not seem to exist at {}".format(
@@ -249,10 +253,19 @@ def get_moe_di_stefano_dataset(options, verbosity=0):
             # Read input data and Clean up the data if there are white spaces around the keys
             with open(options["file"], "r") as data_filehandle:
                 datafile_data = data_filehandle.read()
-            datafile_data = datafile_data.replace('" ', '"')
-            datafile_data = datafile_data.replace(' "', '"')
-            datafile_data = datafile_data.replace(' "', '"')
-            json_data = json.loads(datafile_data)
+                datafile_data = datafile_data.replace('" ', '"')
+                datafile_data = datafile_data.replace(' "', '"')
+                datafile_data = datafile_data.replace(' "', '"')
+                json_data = json.loads(datafile_data)
+
+    if not json_data:
+        # no JSON data or filename given, use the default 2017 dataset
+        verbose_print(
+            "Using the default Moe and de Stefano 2017 datafile",
+            verbosity,
+            1,
+        )
+        json_data = copy.deepcopy(moe_di_stefano_2017_data.moe_di_stefano_2017_data)
 
     return json_data
 
@@ -263,6 +276,51 @@ def imports():
             yield val.__name__
 
 
+def convfloat(x):
+    """
+    Convert scalar x to a float if we can, in which case return the float, otherwise just return x without changing it. Usually, x is a string, but could be anything that float() can handle without failure.
+    """
+    try:
+        y = float(x)
+        return y
+    except ValueError:
+        return x
+
+
+def datalinedict(line: str, parameters: list):
+    """
+    Convert a line of data to a more convenient dictionary.
+    Arguments:
+       line = a line of data as a string
+       parameters = a list of the parameter names
+
+    Note: if the parameter is a floating point number, it will be converted to Python's float type.
+    """
+
+    return {param: convfloat(value) for param, value in zip(parameters, line.split())}
+
+
+def pad_output_distribution(dist: dict, binwidth: float):
+    """
+    Given a distribution, dist (a dictionary), which should be binned every binwidth (float), fill the distribution with zeros when there is no data. Note: this changes the data in place.
+    """
+
+    # sorted list of the keys
+    skeys = sorted(dist.keys(), key=lambda x: float(x))
+
+    # get min and max, offset by the binwidth
+    min_val = skeys[0] - binwidth
+    max_val = skeys[-1] + binwidth
+
+    # pad with zeros
+    x = min_val
+    while x <= max_val:
+        dist[x] = dist.setdefault(x, 0.0)
+        x += binwidth
+
+    return dist
+
+
 class catchtime(object):
     """
     Context manager to calculate time spent
@@ -307,6 +365,7 @@ class Capturing(list):
         del self._stringio  # free up some memory
         sys.stdout = self._stdout
 
+
 def call_binary_c_config(argument):
     """
     Function to interface with the binary_c config file
@@ -328,10 +387,9 @@ def call_binary_c_config(argument):
         msg = "binary_c-config file does not exist. Aborting"
         raise ValueError(msg)
 
-    output = (
-        subprocess.run([BINARY_C_CONFIG, argument], stdout=subprocess.PIPE, check=True)
-        .stdout.decode("utf-8")
-    )
+    output = subprocess.run(
+        [BINARY_C_CONFIG, argument], stdout=subprocess.PIPE, check=True
+    ).stdout.decode("utf-8")
 
     return output
 
@@ -356,6 +414,7 @@ def verbose_print(message: str, verbosity: int, minimal_verbosity: int) -> None:
 
     if verbosity >= minimal_verbosity:
         print(message)
+        sys.stdout.flush()
 
 
 def remove_file(file: str, verbosity: int = 0) -> None:
@@ -391,6 +450,10 @@ def remove_file(file: str, verbosity: int = 0) -> None:
         )
 
 
+def get_username():
+    return psutil.Process().username()
+
+
 def temp_dir(*args: str) -> str:
     """
     Function to create directory within the TMP directory of the file system
@@ -405,7 +468,8 @@ def temp_dir(*args: str) -> str:
     """
 
     tmp_dir = tempfile.gettempdir()
-    path = os.path.join(tmp_dir, "binary_c_python")
+    username = get_username()
+    path = os.path.join(tmp_dir, "binary_c_python-{}".format(username))
 
     # loop over the other paths if there are any:
     if args:
@@ -646,11 +710,21 @@ def parse_binary_c_version_info(version_info_string: str) -> dict:
     cleaned = cleaned - ensembles
 
     ensemble_dict = {}
+    ensemble_filter_dict = {}
     for el in ensembles:
         split_info = el.split("Ensemble ")[-1].split(" is ")
+
         if len(split_info) > 1:
-            ensemble_dict[int(split_info[0])] = split_info[-1]
+            if not split_info[0].startswith("filter"):
+                ensemble_dict[int(split_info[0])] = split_info[-1]
+            else:
+                filter_no = int(split_info[0].replace("filter ", ""))
+                ensemble_filter_dict[filter_no] = split_info[-1]
+
     version_info_dict["ensembles"] = ensemble_dict if ensemble_dict else None
+    version_info_dict["ensemble_filters"] = (
+        ensemble_filter_dict if ensemble_filter_dict else None
+    )
 
     ##########################
     # macros:
@@ -837,7 +911,7 @@ def example_parse_output(output: str, selected_header: str) -> dict:
 
     Args:
         output: binary_c output string
-        selected_header: string header of the output (the start of the line that you want to 
+        selected_header: string header of the output (the start of the line that you want to
             process)
 
     Returns:
@@ -879,14 +953,16 @@ def example_parse_output(output: str, selected_header: str) -> dict:
 
     if len(value_dicts) == 0:
         print(
-            "Sorry, didn't find any line matching your header {}".format(selected_header)
+            "Sorry, didn't find any line matching your header {}".format(
+                selected_header
+            )
         )
         return None
 
     keys = value_dicts[0].keys()
 
     # Construct final dict.
-    final_values_dict = defaultdict(list)
+    final_values_dict = collections.defaultdict(list)
     for value_dict in value_dicts:
         for key in keys:
             final_values_dict[key].append(value_dict[key])
@@ -952,15 +1028,7 @@ def filter_arg_dict(arg_dict: dict) -> dict:
         filtered dictionary (pairs with NULL and Function values are removed)
     """
 
-    old_dict = arg_dict.copy()
-    new_dict = {}
-
-    for key in old_dict.keys():
-        if not old_dict[key] in ["NULL", "Function"]:
-            if not old_dict[key] == "":
-                new_dict[key] = old_dict[key]
-
-    return new_dict
+    return filter_dict_through_values(arg_dict.copy(), ["NULL", "Function", ""])
 
 
 def create_arg_string(
@@ -1434,537 +1502,3 @@ def load_logfile(logfile: str) -> None:
         event_list.append(" ".join(split_line[9:]))
 
     print(event_list)
-
-
-########################################################
-# Ensemble dict functions
-########################################################
-
-
-def inspect_dict(
-    input_dict: dict, indent: int = 0, print_structure: bool = True
-) -> dict:
-    """
-    Function to (recursively) inspect a (nested) dictionary.
-    The object that is returned is a dictionary containing the key of the input_dict, but as value
-    it will return the type of what the value would be in the input_dict
-
-    In this way we inspect the structure of these dictionaries, rather than the exact contents.
-
-    Args:
-        input_dict: dictionary you want to inspect
-        print_structure: (optional, default = True)
-        indent: (optional, default = 0) indent of the first output
-
-    Returns:
-        Dictionary that has the same structure as the input_dict, but as values it has the
-            type(input_dict[key]) (except if the value is a dict)
-    """
-
-    structure_dict = OrderedDict() # TODO: check if this still works
-
-    #
-    for key, value in input_dict.items():
-        structure_dict[key] = type(value)
-
-        if print_structure:
-            print("\t" * indent, key, type(value))
-
-        if isinstance(value, dict):
-            structure_dict[key] = inspect_dict(
-                value, indent=indent + 1, print_structure=print_structure
-            )
-
-    return structure_dict
-
-
-def count_keys_recursive(input_dict):
-    """
-    Function to count the total amount of keys in a dictionary
-    """
-
-    local_count = 0
-    for key in input_dict.keys():
-        local_count += 1
-        if isinstance(input_dict[key], (dict, OrderedDict)):
-            local_count += count_keys_recursive(input_dict[key])
-    return local_count
-
-
-def merge_dicts(dict_1: dict, dict_2: dict) -> dict:
-    """
-    Function to merge two dictionaries in a custom way.
-
-    Behaviour:
-
-    When dict keys are only present in one of either:
-        - we just add the content to the new dict
-
-    When dict keys are present in both, we decide based on the value types how to combine them:
-        - dictionaries will be merged by calling recursively calling this function again
-        - numbers will be added
-        - (opt) lists will be appended
-        - In the case that the instances do not match: for now I will raise an error
-
-    Args:
-        dict_1: first dictionary
-        dict_2: second dictionary
-
-    Returns:
-        Merged dictionary
-
-    """
-
-    # Set up new dict
-    new_dict = OrderedDict() # TODO: check if this still works
-
-    #
-    keys_1 = dict_1.keys()
-    keys_2 = dict_2.keys()
-
-    # Find overlapping keys of both dicts
-    overlapping_keys = set(keys_1).intersection(set(keys_2))
-
-    # Find the keys that are unique
-    unique_to_dict_1 = set(keys_1).difference(set(keys_2))
-    unique_to_dict_2 = set(keys_2).difference(set(keys_1))
-
-    # Add the unique keys to the new dict
-    for key in unique_to_dict_1:
-        # If these items are ints or floats, then just put them in
-        if isinstance(dict_1[key], (float, int)):
-            new_dict[key] = dict_1[key]
-        # Else, to be safe we should deepcopy them
-        else:
-            copy_dict = copy.deepcopy(dict_1[key])
-            new_dict[key] = copy_dict
-
-    for key in unique_to_dict_2:
-        # If these items are ints or floats, then just put them in
-        if isinstance(dict_2[key], (float, int)):
-            new_dict[key] = dict_2[key]
-        # Else, to be safe we should deepcopy them
-        else:
-            copy_dict = copy.deepcopy(dict_2[key])
-            new_dict[key] = copy_dict
-
-    # Go over the common keys:
-    for key in overlapping_keys:
-
-        # See whether the types are actually the same
-        if not type(dict_1[key]) is type(dict_2[key]):
-            # Exceptions:
-            if (type(dict_1[key]) in [int, float, np.float64]) and (
-                type(dict_2[key]) in [int, float, np.float64]
-            ):
-                new_dict[key] = dict_1[key] + dict_2[key]
-            # Exceptions:
-            elif (type(dict_1[key]) in [dict, OrderedDict]) and (
-                type(dict_2[key]) in [dict, OrderedDict]
-            ):
-                new_dict[key] = merge_dicts(dict_1[key], dict_2[key])
-
-            else:
-                print(
-                    "Error key: {} value: {} type: {} and key: {} value: {} type: {} are not of the same type and cannot be merged".format(
-                        key,
-                        dict_1[key],
-                        type(dict_1[key]),
-                        key,
-                        dict_2[key],
-                        type(dict_2[key]),
-                    )
-                )
-                raise ValueError
-
-        # Here we check for the cases that we want to explicitly catch. Ints will be added,
-        # floats will be added, lists will be appended (though that might change) and dicts will be
-        # dealt with by calling this function again.
-        else:
-            # ints
-            # Booleans (has to be the type Bool, not just a 0 or 1)
-            if isinstance(dict_1[key], bool) and isinstance(dict_2[key], bool):
-                new_dict[key] = dict_1[key] or dict_2[key]
-
-            elif isinstance(dict_1[key], int) and isinstance(dict_2[key], int):
-                new_dict[key] = dict_1[key] + dict_2[key]
-
-            # floats
-            elif isinstance(dict_1[key], float) and isinstance(dict_2[key], float):
-                new_dict[key] = dict_1[key] + dict_2[key]
-
-            # lists
-            elif isinstance(dict_1[key], list) and isinstance(dict_2[key], list):
-                new_dict[key] = dict_1[key] + dict_2[key]
-
-            # dicts
-            elif isinstance(dict_1[key], dict) and isinstance(dict_2[key], dict):
-                new_dict[key] = merge_dicts(dict_1[key], dict_2[key])
-
-            else:
-                print(
-                    "Object types {}: {} ({}), {} ({}) not supported.".format(
-                        key,
-                        dict_1[key],
-                        type(dict_1[key]),
-                        dict_2[key],
-                        type(dict_2[key]),
-                    )
-                )
-                raise ValueError
-
-    #
-    return new_dict
-
-
-def update_dicts(dict_1: dict, dict_2: dict) -> dict:
-    """
-    Function to update dict_1 with values of dict_2 in a recursive way.
-
-    Behaviour:
-
-    When dict keys are only present in one of either:
-        - we just add the content to the new dict
-
-    When dict keys are present in both, we decide based on the value types how to combine them:
-        - value of dict2 will be taken
-
-    Args:
-        dict_1: first dictionary
-        dict_2: second dictionary
-
-    Returns:
-        New dictionary with Updated values
-
-    """
-
-    # Set up new dict
-    new_dict = OrderedDict() # TODO: check if this still works
-
-    #
-    keys_1 = dict_1.keys()
-    keys_2 = dict_2.keys()
-
-    # Find overlapping keys of both dicts
-    overlapping_keys = set(keys_1).intersection(set(keys_2))
-
-    # Find the keys that are unique
-    unique_to_dict_1 = set(keys_1).difference(set(keys_2))
-    unique_to_dict_2 = set(keys_2).difference(set(keys_1))
-
-    # Add the unique keys to the new dict
-    for key in unique_to_dict_1:
-        # If these items are ints or floats, then just put them in
-        if isinstance(dict_1[key], (float, int)):
-            new_dict[key] = dict_1[key]
-        # Else, to be safe we should deepcopy them
-        else:
-            copy_dict = copy.deepcopy(dict_1[key])
-            new_dict[key] = copy_dict
-
-    for key in unique_to_dict_2:
-        # If these items are ints or floats, then just put them in
-        if isinstance(dict_2[key], (float, int)):
-            new_dict[key] = dict_2[key]
-        # Else, to be safe we should deepcopy them
-        else:
-            copy_dict = copy.deepcopy(dict_2[key])
-            new_dict[key] = copy_dict
-
-    # Go over the common keys:
-    for key in overlapping_keys:
-
-        # See whether the types are actually the same
-        if not type(dict_1[key]) is type(dict_2[key]):
-            # Exceptions:
-            if (type(dict_1[key]) in [int, float]) and (
-                type(dict_2[key]) in [int, float]
-            ):
-                new_dict[key] = dict_2[key]
-
-            else:
-                print(
-                    "Error key: {} value: {} type: {} and key: {} value: {} type: {} are not of the same type and cannot be merged".format(
-                        key,
-                        dict_1[key],
-                        type(dict_1[key]),
-                        key,
-                        dict_2[key],
-                        type(dict_2[key]),
-                    )
-                )
-                raise ValueError
-
-        # Here we check for the cases that we want to explicitly catch. Ints will be added,
-        # floats will be added, lists will be appended (though that might change) and dicts will be
-        # dealt with by calling this function again.
-        else:
-            # dicts
-            if isinstance(dict_1[key], dict) and isinstance(dict_2[key], dict):
-                new_dict[key] = update_dicts(dict_1[key], dict_2[key])
-            else:
-                new_dict[key] = dict_2[key]
-
-    #
-    return new_dict
-
-
-def multiply_values_dict(input_dict, factor):
-    """
-    Function that goes over dictionary recursively and multiplies the value if possible by a factor
-
-    If the key equals "general_info", the multiplication gets skipped
-    """
-
-    for key in input_dict:
-        if not key == 'general_info':
-            if isinstance(input_dict[key], (dict, OrderedDict)):
-                input_dict[key] = multiply_values_dict(input_dict[key], factor)
-            else:
-                if isinstance(input_dict[key], (int, float)):
-                    input_dict[key] = input_dict[key] * factor
-
-    return input_dict
-
-def custom_sort_dict(input_dict):
-    """
-    Returns a dictionary that is ordered, but can handle numbers better than normal OrderedDict
-
-    When the keys of the current dictionary are of mixed type, we first find all the unique types.
-    Sort that list of type names. Then find the values that fit that type.
-    Sort those and append them to the sorted keys list. 
-    This is done until all the keys are sorted.
-
-    All objects other than dictionary types are directly return as they are
-    """
-
-    # If the new input is a dictionary, then try to sort it
-    if isinstance(input_dict, (dict, OrderedDict)):
-        new_dict = OrderedDict()
-
-        keys = input_dict.keys()
-
-        # Check if types are the same
-        all_types_keys = []
-        for key in keys:
-            if not type(key) in all_types_keys:
-                all_types_keys.append(type(key))
-
-        # If there are multiple types, then we loop over them and do a piece wise sort
-        if len(all_types_keys) > 1:
-            msg = "Different types in the same dictionary key set"
-            
-            # Create a string repr of the type name to sort them afterwards
-            str_types = {repr(el):el for el in all_types_keys}
-
-            # Set up sorted keys list
-            sorted_keys = []
-
-            for key_str_type in sorted(str_types.keys()):
-                cur_type = str_types[key_str_type]
-
-                cur_list = [key for key in keys if isinstance(key, cur_type)]
-                cur_sorted_list = sorted(cur_list)
-
-                sorted_keys = sorted_keys + cur_sorted_list
-        else:
-            sorted_keys = sorted(keys)
-
-        for key in sorted_keys:
-            new_dict[key] = custom_sort_dict(copy.deepcopy(input_dict[key]))
-
-        return new_dict
-    return input_dict
-
-
-def recursive_change_key_to_float(input_dict):
-    """
-    Function to recursively change the key to float
-
-    This only works if the dict contains just sub-dicts or numbers/strings.
-    Does not work with lists as values
-    """
-
-    new_dict = OrderedDict() # TODO: check if this still works
-
-    for key in input_dict:
-        if isinstance(input_dict[key], (dict, OrderedDict)):
-            try:
-                num_key = float(key)
-                new_dict[num_key] = recursive_change_key_to_float(copy.deepcopy(input_dict[key]))
-            except ValueError:
-                new_dict[key] = recursive_change_key_to_float(copy.deepcopy(input_dict[key]))
-        else:
-            try:
-                num_key = float(key)
-                new_dict[num_key] = input_dict[key]
-            except ValueError:
-                new_dict[key] = input_dict[key]
-
-    return new_dict
-
-
-def recursive_change_key_to_string(input_dict):
-    """
-    Function to recursively change the key back to a string but this time in a format that we decide
-    """
-
-    new_dict = OrderedDict() # TODO: check if this still works
-
-    for key in input_dict:
-        if isinstance(input_dict[key], (dict, OrderedDict)):
-            if isinstance(key, (int, float)):
-                string_key = "{:g}".format(key)
-                new_dict[string_key] = recursive_change_key_to_string(copy.deepcopy(input_dict[key]))
-            else:
-                new_dict[key] = recursive_change_key_to_string(copy.deepcopy(input_dict[key]))
-        else:
-            if isinstance(key, (int, float)):
-                string_key = "{:g}".format(key)
-                new_dict[string_key] = input_dict[key]
-            else:
-                new_dict[key] = input_dict[key]
-
-    return new_dict
-
-
-#####
-def extract_ensemble_json_from_string(binary_c_output: str) -> dict:
-    """
-    Function to extract the ensemble_json information from a raw binary_c output string
-
-    Args:
-        binary_c_output: raw binary_c output string
-
-    Returns:
-        JSON dictionary with the parsed ENSEMBLE_JSON data
-    """
-
-    json_dict = None
-
-    try:
-        # If there is no output just return an empty dict:
-        if not binary_c_output:
-            json_dict = {}
-            return json_dict
-
-        ensemble_jsons_strings = [
-            line
-            for line in binary_c_output.splitlines()
-            if line.startswith("ENSEMBLE_JSON")
-        ]
-
-        json_dict = handle_ensemble_string_to_json(
-            ensemble_jsons_strings[0][len("ENSEMBLE_JSON ") :]
-        )
-
-        if len(ensemble_jsons_strings) > 1:
-            verbose_print(
-                "Warning: There is more than one line starting with ENSEMBLE_JSON. Taking the first, but you should check this out.",
-                1,
-                0,
-            )
-    except IndexError:
-        verbose_print(
-            "Error: Couldn't extract the ensemble information from the output string",
-            0,
-            0,
-        )
-
-    return json_dict
-
-
-def handle_ensemble_string_to_json(raw_output):
-    """
-    Function that deals with the raw output of the ensemble and
-    creates a working JSON dictionary out of it.
-
-    Having this wrapper makes it easy to
-
-    Args:
-        raw_output: raw output of the ensemble dump by binary_c
-
-    Returns:
-        json.loads(raw_output, cls=binarycDecoder)
-
-    """
-
-    # return json.loads(json.dumps(ast.literal_eval(raw_output)), cls=binarycDecoder)
-    return json.loads(raw_output, cls=binarycDecoder)
-
-
-def binaryc_json_serializer(obj: Any) -> Any:
-    """
-    Custom serialiser for binary_c to use when functions are present in the dictionary
-    that we want to export.
-
-    Function objects will be turned into str representations of themselves
-
-    Args:
-        obj: The object that might not be serialisable
-
-    Returns:
-        Either string representation of object if the object is a function, or the object itself
-    """
-
-    if inspect.isfunction(obj) or isinstance(obj, py_rinterpolate.Rinterpolate):
-        return str(obj)
-    return obj
-
-
-class binarycDecoder(json.JSONDecoder):
-    """
-    Custom decoder to transform the numbers that are strings to actual floats
-    """
-
-    def decode(self, s):
-        """
-        Entry point function for decoding
-        """
-
-        result = super().decode(
-            s
-        )  # result = super(Decoder, self).decode(s) for Python 2.x
-        return self._decode(result)
-
-    def _decode(self, o):
-        """
-        Depending on the type of object is will determine whether to loop over the elements,
-        or try to change the type of the object from string to float
-
-        The try except might be a somewhat rough solution but it catches all cases.
-        """
-
-        # Check if we can turn it into a float
-        # if isinstance(o, str) or isinstance(o, unicode):
-        if isinstance(o, str):
-            try:
-                return float(o)
-            except ValueError:
-                return o
-        elif isinstance(o, dict):
-            return {k: self._decode(v) for k, v in o.items()}
-        elif isinstance(o, list):
-            return [self._decode(v) for v in o]
-        else:
-            return o
-
-
-class BinaryCEncoder(json.JSONEncoder):
-    """
-    Encoding class function to attempt to convert things to strings.
-    """
-
-    def default(self, o):
-        """
-        Converting function. Well, could be more precise. look at the JSON module
-        """
-        try:
-            str_repr = str(o)
-        except TypeError:
-            pass
-        else:
-            return str_repr
-
-        # Let the base class default method raise the TypeError
-        return json.JSONEncoder.default(self, o)
