@@ -1,14 +1,14 @@
 """
-Module that contains the default options for the population grid code along with the description for these options, in the form of dictionaries: 
+Module that contains the default options for the population grid code along with the description for these options, in the form of dictionaries:
     - grid_options_defaults_dict: dictionary containing the default values for all the options
     - grid_options_descriptions: dictionary containing the description for these options.
 
 There are several other functions in this module, mostly to generate help texts or documents:
-    - grid_options_help: interactive function for the user to get descriptions for options 
+    - grid_options_help: interactive function for the user to get descriptions for options
     - grid_options_description_checker: function that checks that checks which options have a description.
     - write_grid_options_to_rst_file: function to generate the .rst document for the docs
 
-With this its also possible to automatically generate a document containing all the setting names + descriptions. 
+With this its also possible to automatically generate a document containing all the setting names + descriptions.
 
 All the options starting with _ should not be changed by the user except when you really know what you're doing (which is probably hacking the code :P)
 """
@@ -20,29 +20,32 @@ from binarycpython.utils.functions import return_binary_c_version_info
 
 _LOGGER_VERBOSITY_LEVEL = 1
 _CUSTOM_LOGGING_VERBOSITY_LEVEL = 2
-_MS_VERBOSITY_LEVEL = 5
-_MS_VERBOSITY_INTERPOLATOR_LEVEL = 6
-_MS_VERBOSITY_INTERPOLATOR_EXTRA_LEVEL = 7
+_MOE2017_VERBOSITY_LEVEL = 5
+_MOE2017_VERBOSITY_INTERPOLATOR_LEVEL = 6
+_MOE2017_VERBOSITY_INTERPOLATOR_EXTRA_LEVEL = 7
 
 # Options dict
 grid_options_defaults_dict = {
     ##########################
     # general (or unordered..)
     ##########################
-    "amt_cores": 1,  # total amount of cores used to evolve the population
+    "num_cores": 1,  # total number of cores used to evolve the population
     "parse_function": None,  # Function to parse the output with.
-    "multiplicity_fraction_function": 0, # Which multiplicity fraction function to use. 0: None, 1: Arenou 2010, 2: Rhagavan 2010, 3: M&S 2017
+    "multiplicity_fraction_function": 0,  # Which multiplicity fraction function to use. 0: None, 1: Arenou 2010, 2: Rhagavan 2010, 3: Moe and di Stefano 2017
     "tmp_dir": temp_dir(),  # Setting the temp dir of the program
     "_main_pid": -1,  # Placeholder for the main process id of the run.
+    "save_ensemble_chunks": True,  # Force the ensemble chunk to be saved even if we are joining a thread (just in case the joining fails)
     "combine_ensemble_with_thread_joining": True,  # Flag on whether to combine everything and return it to the user or if false: write it to data_dir/ensemble_output_{population_id}_{thread_id}.json
     "_commandline_input": "",
     "log_runtime_systems": 0,  # whether to log the runtime of the systems (1 file per thread. stored in the tmp_dir)
     "_actually_evolve_system": True,  # Whether to actually evolve the systems of just act as if. for testing. used in _process_run_population_grid
     "max_queue_size": 1000,  # Maximum size of the system call queue.
-    "run_zero_probability_system": True, # Whether to run the zero probability systems
+    "run_zero_probability_system": True,  # Whether to run the zero probability systems
     "_zero_prob_stars_skipped": 0,
-    "ensemble_factor_in_probability_weighted_mass": False, # Whether to multiply the ensemble results by 1/probability_weighted_mass
-    "do_dry_run": True, # Whether to do a dry run to calculate the total probability for this run
+    "ensemble_factor_in_probability_weighted_mass": False,  # Whether to multiply the ensemble results by 1/probability_weighted_mass
+    "do_dry_run": True,  # Whether to do a dry run to calculate the total probability for this run
+    "custom_generator": None,  # Place for the custom system generator
+    "exit_after_dry_run": False,  # Exit after dry run?
     ##########################
     # Execution log:
     ##########################
@@ -50,6 +53,8 @@ grid_options_defaults_dict = {
     "log_file": os.path.join(
         temp_dir(), "binary_c_python.log"
     ),  # Set to None to not log to file. The directory will be created
+    "log_dt": 5,  # time between vb=1 logging outputs
+    "n_logging_stats": 50,  # number of logging stats used to calculate time remaining (etc.) default = 50
     ##########################
     # binary_c files
     ##########################
@@ -64,11 +69,12 @@ grid_options_defaults_dict = {
     ),  # TODO: make this more robust
     "_binary_c_dir": os.environ["BINARY_C"],
     ##########################
-    # M&S internal settings
+    # Moe and di Stefano (2017) internal settings
     ##########################
-    "_loaded_ms_data": False, # Holds flag whether the M&S data is loaded into memory
-    "_set_ms_grid": False, # Whether the M&S grid has been loaded
-    "m&s_options": None, # Holds the M&S options.
+    "_loaded_Moe2017_data": False,  # Holds flag whether the Moe and di Stefano (2017) data is loaded into memory
+    "_set_Moe2017_grid": False,  # Whether the Moe and di Stefano (2017) grid has been loaded
+    "Moe2017_options": None,  # Holds the Moe and di Stefano (2017) options.
+    "_Moe2017_JSON_data": None,  # Stores the data
     ##########################
     # Custom logging
     ##########################
@@ -95,6 +101,7 @@ grid_options_defaults_dict = {
     "evolution_type": "grid",  # Flag for type of population evolution
     "_evolution_type_options": [
         "grid",
+        "custom_generator",
     ],  # available choices for type of population evolution. # TODO: fill later with Monte Carlo, source file
     "_system_generator": None,  # value that holds the function that generates the system
     # (result of building the grid script)
@@ -107,18 +114,20 @@ grid_options_defaults_dict = {
     "_start_time_evolution": 0,  # Start time of the grid
     "_end_time_evolution": 0,  # end time of the grid
     "_errors_found": False,  # Flag whether there are any errors from binary_c
-    "_errors_exceeded": False,  # Flag whether the amt of errors have exceeded the limit
-    "_failed_count": 0,  # amt of failed systems
+    "_errors_exceeded": False,  # Flag whether the number of errors have exceeded the limit
+    "_failed_count": 0,  # number of failed systems
     "_failed_prob": 0,  # Summed probability of failed systems
     "failed_systems_threshold": 20,  # Maximum failed systems per process allowed to fail before the process stops logging the failing systems.
     "_failed_systems_error_codes": [],  # List to store the unique error codes
     "_population_id": 0,  # Random id of this grid/population run, Unique code for the population. Should be set only once by the controller process.
     "_total_mass_run": 0,  # To count the total mass that thread/process has ran
     "_total_probability_weighted_mass_run": 0,  # To count the total mass * probability for each system that thread/process has ran
-    "modulo": 1,  # run modulo n of the grid. #TODO: fix this
+    "modulo": 1,  # run modulo n of the grid.
+    "start_at": 0,  # start at the first model
     ## Grid type evolution
     "_grid_variables": {},  # grid variables
     "gridcode_filename": None,  # filename of gridcode
+    "symlink latest gridcode": True,  # symlink to latest gridcode
     ## Monte carlo type evolution
     # TODO: make MC options
     ## Evolution from source file
@@ -157,7 +166,7 @@ grid_options_defaults_dict = {
     # "slurm_jobid": "",  # slurm job id (%A)
     # "slurm_memory": 512,  # in MB, the memory use of the job
     # "slurm_warn_max_memory": 1024,  # in MB : warn if mem req. > this
-    # "slurm_use_all_node_CPUs": 0,  # 1 = use all of a node's CPUs. 0 = use a given amount of CPUs
+    # "slurm_use_all_node_CPUs": 0,  # 1 = use all of a node's CPUs. 0 = use a given number of CPUs
     # "slurm_postpone_join": 0,  # if 1 do not join on slurm, join elsewhere. want to do it off the slurm grid (e.g. with more RAM)
     # "slurm_jobarrayindex": "",  # slurm job array index (%a)
     # "slurm_jobname": "binary_grid",  # default
@@ -447,17 +456,17 @@ grid_options_descriptions = {
     "_binary_c_shared_library": "Full path to the libbinary_c file. This options is not used in the population object",
     "verbosity": "Verbosity of the population code. Default is 0, by which only errors will be printed. Higher values will show more output, which is good for debugging.",
     "binary": "Set this to 1 if the population contains binaries. Input: int",  # TODO: write what effect this has.
-    "amt_cores": "The amount of cores that the population grid will use. The multiprocessing is useful but make sure to figure out how many logical cores the machine has. The core is multi processed, not multi threaded, and will gain no extra speed when amt_cores exceeds the amount of logical cores. Input: int",
+    "num_cores": "The number of cores that the population grid will use. The multiprocessing is useful but make sure to figure out how many logical cores the machine has (use e.g. psutil.cpu_count(logical=False) to find the true number of cores, psutil.cpu_count(logical=True) to find the number of logical cores). The core is multi processed, not multi threaded, and will gain no extra speed when num_cores exceeds the number of logical cores. Input: int",
     "_start_time_evolution": "Variable storing the start timestamp of the population evolution. Set by the object itself.",  # TODO: make sure this is logged to a file
     "_end_time_evolution": "Variable storing the end timestamp of the population evolution. Set by the object itself",  # TODO: make sure this is logged to a file
-    "_total_starcount": "Variable storing the total amount of systems in the generator. Used and set by the population object.",
+    "_total_starcount": "Variable storing the total number of systems in the generator. Used and set by the population object.",
     "_custom_logging_shared_library_file": "filename for the custom_logging shared library. Used and set by the population object",
     "_errors_found": "Variable storing a Boolean flag whether errors by binary_c are encountered.",
-    "_errors_exceeded": "Variable storing a Boolean flag whether the amount of errors was higher than the set threshold (failed_systems_threshold). If True, then the command line arguments of the failing systems will not be stored in the failed_system_log files.",
+    "_errors_exceeded": "Variable storing a Boolean flag whether the number of errors was higher than the set threshold (failed_systems_threshold). If True, then the command line arguments of the failing systems will not be stored in the failed_system_log files.",
     "source_file_filename": "Variable containing the source file containing lines of binary_c command line calls. These all have to start with binary_c.",  # TODO: Expand
     "C_auto_logging": "Dictionary containing parameters to be logged by binary_c. The structure of this dictionary is as follows: the key is used as the headline which the user can then catch. The value at that key is a list of binary_c system parameters (like star[0].mass)",
     "C_logging_code": "Variable to store the exact code that is used for the custom_logging. In this way the user can do more complex logging, as well as putting these logging strings in files.",
-    "_failed_count": "Variable storing the amount of failed systems.",
+    "_failed_count": "Variable storing the number of failed systems.",
     "_evolution_type_options": "List containing the evolution type options.",
     "_failed_prob": "Variable storing the total probability of all the failed systems",
     "_failed_systems_error_codes": "List storing the unique error codes raised by binary_c of the failed systems",
@@ -474,8 +483,8 @@ grid_options_descriptions = {
     "_probtot": "Total probability of the population.",  # TODO: check whether this is used properly throughout
     "_main_pid": "Main process ID of the master process. Used and set by the population object.",
     "_store_memaddr": "Memory address of the store object for binary_c.",
-    "failed_systems_threshold": "Variable storing the maximum amount of systems that are allowed to fail before logging their command line arguments to failed_systems log files",
-    "parse_function": "Function that the user can provide to handle the output the binary_c. This function has to take the arguments (self, output). Its best not to return anything in this function, and just store stuff in the grid_options['results'] dictionary, or just output results to a file",
+    "failed_systems_threshold": "Variable storing the maximum number of systems that are allowed to fail before logging their command line arguments to failed_systems log files",
+    "parse_function": "Function that the user can provide to handle the output the binary_c. This function has to take the arguments (self, output). Its best not to return anything in this function, and just store stuff in the self.grid_results dictionary, or just output results to a file",
     "condor": "Int flag whether to use a condor type population evolution. Not implemented yet.",  # TODO: describe this in more detail
     "slurm": "Int flag whether to use a Slurm type population evolution.",  # TODO: describe this in more detail
     "weight": "Weight factor for each system. The calculated probability is multiplied by this. If the user wants each system to be repeated several times, then this variable should not be changed, rather change the _repeat variable instead, as that handles the reduction in probability per system. This is useful for systems that have a process with some random element in it.",  # TODO: add more info here, regarding the evolution splitting.
@@ -487,21 +496,31 @@ grid_options_descriptions = {
     "_total_probability_weighted_mass_run": "To count the total mass * probability for each system that thread/process has ran",
     "_actually_evolve_system": "Whether to actually evolve the systems of just act as if. for testing. used in _process_run_population_grid",
     "max_queue_size": "Maximum size of the queue that is used to feed the processes. Don't make this too big! Default: 1000. Input: int",
-    "_set_ms_grid": "Internal flag whether the M&S grid has been loaded",
+    "_set_Moe2017_grid": "Internal flag whether the Moe and di Stefano (2017) grid has been loaded",
     "run_zero_probability_system": "Whether to run the zero probability systems. Default: True. Input: Boolean",
     "_zero_prob_stars_skipped": "Internal counter to track how many systems are skipped because they have 0 probability",
     "ensemble_factor_in_probability_weighted_mass": "Flag to multiply all the ensemble results with 1/probability_weighted_mass",
-    "multiplicity_fraction_function": "Which multiplicity fraction function to use. 0: None, 1: Arenou 2010, 2: Rhagavan 2010, 3: M&S 2017",
-    "m&s_options": "Internal variable that holds the M&S options. Don't write to this your self",
-    "_loaded_ms_data": "Internal variable storing whether the M&S data has been loaded into memory",
+    "multiplicity_fraction_function": "Which multiplicity fraction function to use. 0: None, 1: Arenou 2010, 2: Rhagavan 2010, 3: Moe and di Stefano (2017) 2017",
+    "m&s_options": "Internal variable that holds the Moe and di Stefano (2017) options. Don't write to this your self",
+    "_loaded_Moe2017_data": "Internal variable storing whether the Moe and di Stefano (2017) data has been loaded into memory",
     "do_dry_run": "Whether to do a dry run to calculate the total probability for this run",
+    "_Moe2017_JSON_data": "Location to store the loaded Moe&diStefano2017 dataset",  # Stores the data
 }
 
 ###
 #
 
+MIN_MASS_BINARY_C = float(
+    return_binary_c_version_info(parsed=True)["macros"]["BINARY_C_MINIMUM_STELLAR_MASS"]
+)
+
+
 # Default options for the Moe & di Stefano grid
 moe_di_stefano_default_options = {
+    # place holder for the JSON data to be used if a file
+    # isn't specified
+    "JSON": None,
+    # resolution data
     "resolutions": {
         "M": [
             20,  # M1
@@ -520,10 +539,16 @@ moe_di_stefano_default_options = {
             0,  # e3 (quadruple eccentricity) currently unused
         ],
     },
+    "samplerfuncs": {
+        "M": [None, None, None, None],
+        "logP": [None, None, None],
+        "ecc": [None, None, None],
+    },
     "ranges": {
         # stellar masses (Msun)
         "M": [
-            0.08,  # 0.08 is a tad bit above the minimum mass. Don't sample at 0.07, otherwise the first row of q values will have a phasevol of 0. Anything higher is fine.
+            MIN_MASS_BINARY_C
+            * 1.05,  # 0.08 is a tad bit above the minimum mass. Don't sample at 0.07, otherwise the first row of q values will have a phasevol of 0. Anything higher is fine.
             80.0,  # (rather arbitrary) upper mass cutoff
         ],
         "q": [
@@ -534,8 +559,7 @@ moe_di_stefano_default_options = {
         "ecc": [0.0, 0.99],
     },
     # minimum stellar mass
-    "Mmin": float(return_binary_c_version_info(parsed=True)['macros']['BINARY_C_MINIMUM_STELLAR_MASS']), # We take the value that binary_c has set as the default 
-
+    "Mmin": MIN_MASS_BINARY_C,  # We take the value that binary_c has set as the default
     # multiplicity model (as a function of log10M1)
     #
     # You can use 'Poisson' which uses the system multiplicity
@@ -546,7 +570,6 @@ moe_di_stefano_default_options = {
     # from the data, but then triples and quadruples are
     # combined (and there are NO quadruples).
     "multiplicity_model": "Poisson",
-
     # multiplicity modulator:
     # [single, binary, triple, quadruple]
     #
@@ -709,7 +732,7 @@ def grid_options_description_checker(print_info: bool = True) -> int:
         print_info: whether to print out information about which options contain proper descriptions and which do not
 
     Returns:
-        the amount of undescribed keys
+        the number of undescribed keys
     """
 
     # Get the keys
@@ -814,7 +837,9 @@ def print_option_descriptions(filehandle, options, descriptions, title, extra_te
     for option in sorted(options):
         if option in descriptions:
             print(
-                "| **{}**: {}".format(option, descriptions[option].replace("\n", "\n\t")),
+                "| **{}**: {}".format(
+                    option, descriptions[option].replace("\n", "\n\t")
+                ),
                 file=filehandle,
             )
         else:
