@@ -344,7 +344,7 @@ def plot_masses(df, show_stellar_types: bool = False, show_plot: bool = True):
     )
 
     # Mass 1
-    fig.axes[0].plot(df["time"], df["mass_1"], label="star 1", color='red')
+    fig.axes[0].plot(df["time"], df["mass_1"], label="star 1", color="red")
     fig.axes[0].axhline(
         df["pms_mass_1"].values.tolist()[0],
         color="red",
@@ -354,7 +354,8 @@ def plot_masses(df, show_stellar_types: bool = False, show_plot: bool = True):
         alpha=0.5,
     )
     fig.axes[0].plot(
-        df["time"], df["core_mass_1"],
+        df["time"],
+        df["core_mass_1"],
         color="red",
         linestyle="-.",
         linewidth=2,
@@ -373,7 +374,8 @@ def plot_masses(df, show_stellar_types: bool = False, show_plot: bool = True):
         alpha=0.5,
     )
     fig.axes[0].plot(
-        df["time"], df["core_mass_2"],
+        df["time"],
+        df["core_mass_2"],
         color="orange",
         linestyle="-.",
         linewidth=2,
